/*
 *  polyFactor.cpp
 *  fp
 *
 *  Created by Robert Delaney on 6/4/13.
 *  Copyright 2013 Bob Delaney's Science Software. All rights reserved.
 *
 */

#include "fp.h"
#include "poly.h"
#include "polyMath.h"
#include "matq.h"
#include "matqMath.h"
#include "polyFactor.h"


// Q is a square matrix of dimension equal to the degree of the polynomial a(x)
void polyCalcQmatrix(matq& Q, const poly& a)
{
	INT32		i, k;
	poly		xqk, h;
	
	Q.nr = a.deg;
	Q.nc = a.deg;
	
	Q.array = (bf**)malloc(Q.nr*sizeof(bf*));
	for(i=0;i<Q.nr;i++)
		Q.array[i] = (bf*)malloc(Q.nc*sizeof(bf));
	
	// 0th column is transpose of (1, 0, ... , 0)
	
	Q.array[0][0] = 1;
	for(i=1;i<Q.nr;i++)
		Q.array[i][0] = 0;
	
	// Use pdf: Algorithms for Factoring Square-Free Polynomials over Finite Fields
	xqk.deg = 1;
	xqk.array = (bf*)malloc((xqk.deg+1)*sizeof(bf));
	
	xqk.array[0] = 0;
	xqk.array[1] = 1;
	
	h = polyMod(xqk, a);
	// column 1 of Q are the coefficients of h = x^q mod a(x) from low to high, h being of degree n-1 or less
	for(i=0;i<=h.deg;i++)
		Q.array[i][1] = h.array[i];
	for(i=h.deg+1;i<Q.nr;i++)
		Q.array[i][1] = 0;
	
	xqk = h;
	for(k=2;k<a.deg;k++)
	{
		xqk = polyMod(h*xqk, a);
		for(i=0;i<=xqk.deg;i++)
			Q.array[i][k] = xqk.array[i];
		for(i=xqk.deg+1;i<Q.nr;i++)
			Q.array[i][k] = 0;
	}
	
}/* polyCalcQmatrix */


// Q is a square matrix of dimension equal to the degree of the polynomial a(x)
void polyCalcQmatrix(matq& Q, const poly& a, INT32 q)
{
	INT32		i, k;
	poly		xqk, h;
	double		qq;
	
	qq = q;
	
	Q.nr = a.deg;
	Q.nc = a.deg;
	
	Q.array = (bf**)malloc(Q.nr*sizeof(bf*));
	for(i=0;i<Q.nr;i++)
		Q.array[i] = (bf*)malloc(Q.nc*sizeof(bf));
	
	// 0th column is transpose of (1, 0, ... , 0)
	
	Q.array[0][0] = 1;
	for(i=1;i<Q.nr;i++)
		Q.array[i][0] = 0;
	
	// Use pdf: Algorithms for Factoring Square-Free Polynomials over Finite Fields
	xqk.deg = 1;
	xqk.array = (bf*)malloc((xqk.deg+1)*sizeof(bf));
	
	xqk.array[0] = 0;
	xqk.array[1] = 1;
	
	h = polyPowerMod(xqk, q, a, qq);
	// column 1 of Q are the coefficients of h = x^q mod a(x) from low to high, h being of degree n-1 or less
	for(i=0;i<=h.deg;i++)
		Q.array[i][1] = h.array[i];
	for(i=h.deg+1;i<Q.nr;i++)
		Q.array[i][1] = 0;
	
	xqk = h;
	for(k=2;k<a.deg;k++)
	{
		xqk = polyMod(h*xqk, a, qq);
		for(i=0;i<=xqk.deg;i++)
			Q.array[i][k] = xqk.array[i];
		for(i=xqk.deg+1;i<Q.nr;i++)
			Q.array[i][k] = 0;
	}
	
}/* polyCalcQmatrix */


// puts polys in rows of matrix where numRows=numPolys and numCols=maxPolyDeg+1; for smaller deg fill row end with zeroes for smaller degrees
void polySquareFree2(matq& z, const poly& x)
{
	poly		*z1;
	long		i, j, numPolys;
	
	polySquareFree(z1, numPolys, x);
	
	z.nr = numPolys;
	
	z.nc = 0;
	for(i=0;i<numPolys;i++)
	{
		if(z1[i].deg>z.nc)
			z.nc = z1[i].deg;
	}
	z.nc++;
	
	z.array = (bf**)malloc(z.nr*sizeof(bf*));
	for(i=0;i<z.nr;i++)
		z.array[i] = (bf*)malloc(z.nc*sizeof(bf));
	
	for(i=0;i<numPolys;i++)
	{
		for(j=0;j<z.nc;j++)
		{
			if(j<=z1[i].deg)
				z.array[i][j] = z1[i].array[j];
			else
				z.array[i][j] = 0;
		}
	}
	
}/* polySquareFree2 */


void polySFfindFactor(poly& z, const poly& a, const matq& nullVectors, double q)
{
	poly		v;
	INT32		i, k;
	bf			s;
	
	v.deg = nullVectors.nr - 1;
	v.array = (bf*)malloc((v.deg+1)*sizeof(bf));
	
	// first try 2nd column null vector; worry about high order zeroes
	//k = 1;
	for(k=1;k<nullVectors.nc;k++)
	{
		for(i=0;i<=v.deg;i++)
			v.array[i] = nullVectors.array[i][k];
		
		v = polyNormalize(v);
		//cout << "v = " << v << endl;
		
		for(i=0;i<q;i++)
		{
			s = i;
			z = gcd(a, v-s, q);
			if(z.deg>0)
				break;
		}
		
		//cout << "s = " << s << endl;
		
		if(z.deg>0)
		{
			s = z.array[z.deg];
			if(!(s.num==1))
			{
				invMod(s.num, s.num, q);
				z = s * z;
			}
			z = z % q;
			return;
		}
	}
		
}/* polySFfindFactor */


void polySFfindAllFactors(INT32& numFactors, poly*& factor, const poly& a, const matq& nullVectors, double q)
{
	INT32		i;
	poly		at, zt;
	
	at = a;
	numFactors = nullVectors.nc;
	
	factor = (poly*)malloc(numFactors*sizeof(poly));
	
	if(numFactors==1)
	{
		factor[0] = a;
		return;
	}
	
	for(i=0;i<numFactors;i++)
	{
		polySFfindFactor(zt, at, nullVectors, q);
		factor[i] = zt;
		at = at / zt;
	}

}/* polySFfindAllFactors */

// poly x should have integer coefficients and must be square free
// also the highest degree coefficient should be 1
void polySFfindTrueFactors(INT32& numFactors, poly*& factor, const poly& x)
{
	matq		Q, nullVectors, reduced;
	double		q;
	mb			qmb;
	mb			maxCoeff;
	INT32		i, numNV;
	bool		isGood;
	
	// find max coeff of x in abs value
	maxCoeff = 0;
	for(i=0;i<=x.deg;i++)
	{
		if(abs(x.array[i].num)>maxCoeff)
			maxCoeff = abs(x.array[i].num);
	}
	
	qmb = 2.*maxCoeff + 1.;
	while(!isItPrime(qmb, 35))
		qmb+=2;
	
	equate(q, qmb);
	
	isGood = false;
	while(!isGood)
	{
		polyCalcQmatrix(Q, x, q);
		Q = Q - matqUnit(Q.nr);
		matqGaussElim(Q, Q, q);
		matqNullSpace(numNV, nullVectors, reduced, Q);
		polySFfindAllFactors(numFactors, factor, x, nullVectors, q);
		isGood = true;
	}
	
	
}/* polySFfindTrueFactors */


// puts polys in rows of matrix where numRows=numPolys and numCols=maxPolyDeg+1; for smaller deg fill row end with zeroes for smaller degrees
// poly x should have integer coefficients and must be square free
// also the highest degree coefficient should be 1
void polySFfindTrueFactors2(matq& factorMatrix, const poly& x)
{
	poly		*factor;
	INT32		i, j, numFactors;
	
	polySFfindTrueFactors(numFactors, factor, x);
	factorMatrix.nr = numFactors;
	
	factorMatrix.nc = 0;
	for(i=0;i<numFactors;i++)
	{
		if(factor[i].deg>factorMatrix.nc)
			factorMatrix.nc = factor[i].deg;
	}
	factorMatrix.nc++;
	
	factorMatrix.array = (bf**)malloc(factorMatrix.nr*sizeof(bf*));
	for(i=0;i<factorMatrix.nr;i++)
		factorMatrix.array[i] = (bf*)malloc(factorMatrix.nc*sizeof(bf));
	
	for(i=0;i<numFactors;i++)
	{
		for(j=0;j<factorMatrix.nc;j++)
		{
			if(j<=factor[i].deg)
				factorMatrix.array[i][j] = factor[i].array[j];
			else
				factorMatrix.array[i][j] = 0;
		}
	}
	
}/* polySFfindTrueFactors2 */



void polyFromMatrixRow(poly& z, long theRow, const matq& x)
{
	long		j;
	
	if(theRow<0 || theRow>=x.nr)
		return;
	
	z.deg = x.nc - 1;
	z.array = (bf*)malloc(x.nc*sizeof(bf));

	for(j=0;j<x.nc;j++)
		z.array[j] = x.array[theRow][j];
		
	z = polyNormalize(z);
		
}/* polyFromMatrixRow */


void findPowersOfSquareFreeFactors(matq& z1, long*& theExp, long& numPolys, const poly& x)
{
	poly		*z;
	poly		zt, r, q, temp;
	matq		z1t;
	long		i, j, maxDeg;
	bool		isGood;
	
	polySquareFree(z, numPolys, x);
	
	if(numPolys<=1)
	{
		z1.nr = 1;
		z1.nc = x.deg + 1;
		z1.array = (bf**)malloc(z1.nr*sizeof(bf*));
		for(i=0;i<z1.nr;i++)
			z1.array[i] = (bf*)malloc(z1.nc*sizeof(bf));
		
		for(i=0;i<z1.nr;i++)
			for(j=0;j<z1.nc;j++)
				z1.array[i][j] = z[i].array[j];
		
		theExp = (long*)malloc(sizeof(long));
		theExp[0] = 1;
		return;
	}
	
	theExp = (long*)malloc((numPolys-1)*sizeof(long));
	for(i=0;i<(numPolys-1);i++)
		theExp[i] = 1;
	
	for(i=1;i<numPolys;i++)
	{
		zt = z[0]; //zt contains powers of some or all of z[1], z[2], ...
		isGood = true;
		while(isGood && z[i].deg)
		{
			divEuclid(q, r, zt, z[i]);
			if(!r.deg && !r.array[0])
			{
				theExp[i-1]++;
				zt = q;
			}
			else
			{
				isGood = false;
			}
		}
	}
	
	// find max deg except for 0th poly
	maxDeg = 0;
	for(i=1;i<numPolys;i++)
	{
		if(z[i].deg>maxDeg)
			maxDeg = z[i].deg;
	}
	
	numPolys--;
	z1t.nr = numPolys;
	z1t.nc = maxDeg + 1;
	z1t.array = (bf**)malloc(z1t.nr*sizeof(bf*));
	for(i=0;i<z1t.nr;i++)
		z1t.array[i] = (bf*)malloc(z1t.nc*sizeof(bf));
	
	for(i=0;i<z1t.nr;i++)
	{
		for(j=0;j<z1t.nc;j++)
		{
			if(j<=z[i+1].deg)
				z1t.array[i][j] = z[i+1].array[j];
			else {
				z1t.array[i][j] = 0;
			}
		}
	}
	
	z1 = z1t;
	//free(z);
	
}/* findPowersOfSquareFreeFactors */


void findPowersOfSquareFreeFactors(matq& z, matq& theExpColM, long& numPolys, const poly& x)
{
	long		*theExp, i;
	
	findPowersOfSquareFreeFactors(z, theExp, numPolys, x);
	
	theExpColM.nr = z.nr;
	theExpColM.nc = 1;
	theExpColM.array = (bf**)malloc(theExpColM.nr*sizeof(bf*));
	for(i=0;i<theExpColM.nr;i++)
		theExpColM.array[i] = (bf*)malloc(theExpColM.nc*sizeof(bf));
	
	for(i=0;i<theExpColM.nr;i++)
		theExpColM.array[i][0] = theExp[i];
	
}/* findPowersOfSquareFreeFactors */


